/* KeyboardHost: Copyright 2011 John Elliott <jce@seasip.demon.co.uk>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package info.seasip.keyboardhost;

import java.awt.Color;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JToggleButton;
import javax.swing.SpinnerListModel;
import javax.swing.SpinnerModel;

/** A panel for customising (some) keyboard settings in the hybrid 2k ROM.
 * Intended to be run hosted in a JOptionPane.
 *
 * @author John Elliott
 */
public class Customize1kPanel extends javax.swing.JPanel
{
    // The copy of the ROM we are modifying.
    private byte [] rom;

    // The names of the 127 set 3 scancodes (excluding 00).
    static String [] scanCodes3 =
    {
        "01 L9",
        "83 L7 Print",
        "03 L5 ChgSc / Jump",
        "04 L3 WS / ExSel",
        "05 L1 SysRq",
        "06 L2 Clear / Test",
        "07 PF1",
        "08 PF13",
        "09 L10 / Ctrl",
        "0A L8 / Pause",
        "0B L6 / EraseEOF",
        "0C L4 / Finish",
        "0D Tab",
        "0E `",
        "0F PF2",
        "10 PF14",
        "11 Reset / Quit",
        "12 Left Shift",
        "13 > <",
        "14 Caps Lock",
        "15 Q",
        "16 1",
        "17 PF3",
        "18 PF15",
        "19 Left Alt",
        "1A Z",
        "1B S",
        "1C A",
        "1D W",
        "1E 2",
        "1F PF4",
        "20 PF16",
        "21 C",
        "22 X",
        "23 D",
        "24 E",
        "25 4",
        "26 3",
        "27 PF5",
        "28 PF17",
        "29 Space",
        "2A V",
        "2B F",
        "2C T",
        "2D R",
        "2E 5",
        "2F PF6",
        "30 PF18",
        "31 N",
        "32 B",
        "33 H",
        "34 G",
        "35 Y",
        "36 6",
        "37 PF7",
        "38 PF19",
        "39 Right Alt",
        "3A M",
        "3B J",
        "3C U",
        "3D 7",
        "3E 8",
        "3F PF8",
        "40 PF20",
        "41 ,",
        "42 K",
        "43 I",
        "44 O",
        "45 0",
        "46 9",
        "47 PF9",
        "48 PF21",
        "49 .",
        "4A /",
        "4B L",
        "4C ;",
        "4D P",
        "4E _",
        "4F PF10",
        "50 PF22",
        "51 Under Right Shift",
        "52 @ '",
        "53 # ~",
        "54 [",
        "55 =",
        "56 PF11",
        "57 PF23",
        "58 Right Ctrl",
        "59 Right Shift",
        "5A Return",
        "5B ]",
        "5C # \\",
        "5D Under Backspace",
        "5E PF12",
        "5F PF24",
        "60 Cursor down",
        "61 Cursor left",
        "62 Home",
        "63 Cursor up",
        "64 Backtab",
        "65 Insert",
        "66 Backspace",
        "67 PA1",
        "68 Under keypad 0",
        "69 Keypad 1",
        "6A Cursor right",
        "6B Keypad 4",
        "6C Keypad 7",
        "6D Del",
        "6E PA2",
        "6F PA3",
        "70 Keypad 0",
        "71 Keypad .",
        "72 Keypad 2",
        "73 Keypad 5",
        "74 Keypad 6",
        "75 Keypad 8",
        "76 Escape",
        "77 Num Lock",
        "78 Under Keypad Enter",
        "79 Keypad Enter",
        "7A Keypad 3",
        "7B Keypad -",
        "7C Keypad + / tab",
        "7D Keypad 9",
        "7E Scroll Lock",
        "84 Keypad Space",
    };

    // The names of the 256 possible translations for each scancode.
    static String[] scanCodes2 =
    {
        "00 Buffer full",
        "01 F9",
        "02 --",
        "03 F5",
        "04 F3",
        "05 F1",
        "06 F2",
        "07 F12",
        "08 F17",
        "09 F10",
        "0A F8",
        "0B F6",
        "0C F4",
        "0D Tab",
        "0E `",
        "0F",
        "10 F18",
        "11 Left Alt",
        "12 Left Shift",
        "13",
        "14 Left Ctrl",
        "15 Q",
        "16 1",
        "17 PA1",
        "18 F19",
        "19 Attn",
        "1A Z",
        "1B S",
        "1C A",
        "1D W",
        "1E 2",
        "1F F13",
        "20 F19",
        "21 C",
        "22 X",
        "23 D",
        "24 E",
        "25 4",
        "26 3",
        "27 F14",
        "28 F20",
        "29 Space",
        "2A V",
        "2B F",
        "2C T",
        "2D R",
        "2E 5",
        "2F F15",
        "30 F21",
        "31 N",
        "32 B",
        "33 H",
        "34 G",
        "35 Y",
        "36 6",
        "37",
        "38 F23",
        "39 CrSel",
        "3A M",
        "3B J",
        "3C U",
        "3D 7",
        "3E 8",
        "3F",
        "40 F24",
        "41 ,",
        "42 K",
        "43 I",
        "44 O",
        "45 0",
        "46 9",
        "47",
        "48",
        "49 .",
        "4A /",
        "4B L",
        "4C ;",
        "4D P",
        "4E _",
        "4F",
        "50 ErEOF",
        "51 Under Right Shift",
        "52 @ '",
        "53 ExSel",
        "54 [",
        "55 =",
        "56",
        "57",
        "58 Caps Lock",
        "59 Right Shift",
        "5A Return",
        "5B ]",
        "5C L1",
        "5D # \\",
        "5E F16",
        "5F Clear",
        "60",
        "61 | \\",
        "62",
        "63 Under keypad Enter",
        "64",
        "65",
        "66 Backspace",
        "67",
        "68 Under keypad 0",
        "69 Keypad 1",
        "6A Under backspace",
        "6B Keypad 4",
        "6C Keypad 7",
        "6D Under keypad +",
        "6E",
        "6F Copy",
        "70 Keypad 0",
        "71 Keypad .",
        "72 Keypad 2",
        "73 Keypad 5",
        "74 Keypad 6",
        "75 Keypad 8",
        "76 Escape",
        "77 Num Lock",
        "78 F11",
        "79 Keypad +",
        "7A Keypad 3",
        "7B Keypad -",
        "7C Keypad *",
        "7D Keypad 9",
        "7E Scroll Lock",
        "7F --",
        "80 --",
        "81 --",
        "82 --",
        "83 F7",
        "84 SysRQ",
        "85 --",
        "86 --",
        "87 --",
        "88 --",
        "89 --",
        "8A --",
        "8B --",
        "8C --",
        "8D --",
        "8E --",
        "8F --",
        "90 --",
        "91 --",
        "92 --",
        "93 --",
        "94 --",
        "95 --",
        "96 --",
        "97 --",
        "98 --",
        "99 --",
        "9A --",
        "9B --",
        "9C --",
        "9D --",
        "9E --",
        "9F --",
        "A0 --",
        "A1 --",
        "A2 --",
        "A3 --",
        "A4 --",
        "A5 --",
        "A6 --",
        "A7 --",
        "A8 --",
        "A9 --",
        "AA Self-test passed",
        "AB --",
        "AC --",
        "AD --",
        "AE --",
        "AF --",
        "B0 --",
        "B1 --",
        "B2 --",
        "B3 --",
        "B4 --",
        "B5 --",
        "B6 --",
        "B7 --",
        "B8 --",
        "B9 --",
        "BA --",
        "BB --",
        "BC --",
        "BD --",
        "BE --",
        "BF --",
        "C0 --",
        "C1 --",
        "C2 --",
        "C3 --",
        "C4 --",
        "C5 --",
        "C6 --",
        "C7 --",
        "C8 --",
        "C9 --",
        "CA --",
        "CB --",
        "CC --",
        "CD --",
        "CE --",
        "CF --",
        "E0 extended key",
        "E1 extended key",
        "E2 --",
        "E3 --",
        "E4 --",
        "E5 --",
        "E6 --",
        "E7 --",
        "E8 --",
        "E9 --",
        "EA --",
        "EB --",
        "EC --",
        "ED --",
        "EE echo",
        "EF --",
        "F0 --",
        "F1 --",
        "F2 --",
        "F3 --",
        "F4 --",
        "F5 --",
        "F6 --",
        "F7 --",
        "F8 --",
        "F9 --",
        "FA ACK",
        "FB --",
        "FC Controller failed",
        "FD Keyboard failed",
        "FE NAK",
        "FF --",
    };

    // The key positions of the 127 keys. Very similar to the same table in
    // KeyboardPanel, but omits the dummy key with internal number 0x50.
    private static final int keyPos[] =
    {
        0, 14, 2, 2,    // 01 L9
        0, 12, 2, 2,    // 83 L7
        0, 10, 2, 2,    // 03 L5
        0,  8, 2, 2,    // 04 L3
        0,  6, 2, 2,    // 05 L1
        2,  6, 2, 2,    // 06 L2
        8,  2, 2, 2,    // 07 PF1
        8,  0, 2, 2,    // 08 PF13
        2, 14, 2, 2,    // 09 L10
        2, 12, 2, 2,    // 0A L8
        2, 10, 2, 2,    // 0B L6
        2,  8, 2, 2,    // 0C L5
        6,  8, 3, 2,    // 0D Tab
        6,  6, 2, 2,    // 0E Backtick
        10, 2, 2, 2,    // 0F PF2
        10, 0, 2, 2,    // 10 PF14
        6, 14, 4, 2,    // 11 Ctrl
        6, 12, 3, 2,    // 12 Shift
        9, 12, 2, 2,    // 13 Backslash
        6, 10, 4, 2,    // 14 Caps Lock
        9,  8, 2, 2,    // 15 Q
        8,  6, 2, 2,    // 16 1
        12, 2, 2, 2,    // 17 PF3
        12, 0, 2, 2,    // 18 PF15
        12,14, 3, 2,    // 19 Alt
        11,12, 2, 2,    // 1A Z
        12,10, 2, 2,    // 1B S
        10,10, 2, 2,    // 1C A
        11, 8, 2, 2,    // 1D W
        10, 6, 2, 2,    // 1E 2
        14, 2, 2, 2,    // 1F PF4
        14, 0, 2, 2,    // 20 PF16
        15,12, 2, 2,    // 21 C
        13,12, 2, 2,    // 22 X
        14,10, 2, 2,    // 23 D
        13, 8, 2, 2,    // 24 E
        14, 6, 2, 2,    // 25 4
        12, 6, 2, 2,    // 26 3
        16, 2, 2, 2,    // 27 PF5
        16, 0, 2, 2,    // 28 PF17
        15,14,13, 2,    // 29 Space
        17,12, 2, 2,    // 2A V
        16,10, 2, 2,    // 2B F
        17, 8, 2, 2,    // 2C T
        15, 8, 2, 2,    // 2D R
        16, 6, 2, 2,    // 2E 5
        18, 2, 2, 2,    // 2F PF6
        18, 0, 2, 2,    // 30 PF18
        21,12, 2, 2,    // 31 N
        19,12, 2, 2,    // 32 B
        20,10, 2, 2,    // 33 H
        18,10, 2, 2,    // 34 G
        19, 8, 2, 2,    // 35 Y
        18, 6, 2, 2,    // 36 6
        20, 2, 2, 2,    // 37 PF7
        20, 0, 2, 2,    // 38 PF19
        28,14, 3, 2,    // 39 Right Alt
        23,12, 2, 2,    // 3A M
        22,10, 2, 2,    // 3B J
        21, 8, 2, 2,    // 3C U
        20, 6, 2, 2,    // 3D 7
        22, 6, 2, 2,    // 3E 8
        22, 2, 2, 2,    // 3F PF8
        22, 0, 2, 2,    // 40 PF20
        25,12, 2, 2,    // 41 ,
        24,10, 2, 2,    // 42 K
        23, 8, 2, 2,    // 43 I
        25, 8, 2, 2,    // 44 O
        26, 6, 2, 2,    // 45 0
        24, 6, 2, 2,    // 46 9
        24, 2, 2, 2,    // 47 PF9
        24, 0, 2, 2,    // 48 PF21
        27,12, 2, 2,    // 49 ,
        29,12, 2, 2,    // 4A /
        26,10, 2, 2,    // 4B L
        28,10, 2, 2,    // 4C ;
        27, 8, 2, 2,    // 4D P
        28, 6, 2, 2,    // 4E -
        26, 2, 2, 2,    // 4F PF10
        26, 0, 2, 2,    // 50 PF22
        31,12, 2, 2,    // 51 Under right Shift
        30,10, 2, 2,    // 52 '
        32,10, 2, 2,    // 53 #
        29, 8, 2, 2,    // 54 [
        30, 6, 2, 2,    // 55 =
        28, 2, 2, 2,    // 56 PF11
        28, 0, 2, 2,    // 57 PF23
        33,14, 3, 2,    // 58 Right Ctrl
        33,12, 3, 2,    // 59 Right Shift
        34,10, 2, 2,    // 5A Return
        31, 8, 2, 2,    // 5B ]
        33, 8, 3, 2,    // 5C Other #
        32, 6, 2, 2,    // 5D Under backspace
        30, 2, 2, 2,    // 5E PF12
        30, 0, 2, 2,    // 5F PF24
        40, 14,2, 2,    // 60 Down
        38, 12,2, 2,    // 61 Left
        40, 12,2, 2,    // 62 Home
        40, 10,2, 2,    // 63 Up
        38,  8,2, 2,    // 64 Ins
        40,  8,2, 2,    // 65 PgDn
        34,  6,2, 2,    // 66 Backspace
        38,  6,2, 2,    // 67 PA1
        46, 14,2, 2,    // 68 Under keypad 0
        46, 12,2, 2,    // 69 Keypad 1
        42, 12,2, 2,    // 6A Right
        46, 10,2, 2,    // 6B Keypad 4
        46,  8,2, 2,    // 6C Keypad 7
        42,  8,2, 2,    // 6D Del
        40,  6,2, 2,    // 6E PA2
        42,  6,2, 2,    // 6F PA3
        48, 14,2, 2,    // 70 Keypad 0
        50, 14,2, 2,    // 71 Keypad .
        48, 12,2, 2,    // 72 Keypad 2
        48, 10,2, 2,    // 73 Keypad 5
        50, 10,2, 2,    // 74 Keypad 5
        48,  8,2, 2,    // 75 Keypad 8
        46,  6,2, 2,    // 76 Num lock
        48,  6,2, 2,    // 77 Keypad /
        52, 14,2, 2,    // 78 Under keypad enter
        52, 12,2, 2,    // 79 Keypad enter
        50, 12,2, 2,    // 7A Keypad 3
        52, 10,2, 2,    // 7B Under keypad +
        52,  8,2, 2,    // 7C Keypad +
        50,  8,2, 2,    // 7D Keypad 9
        50,  6,2, 2,    // 7E Keypad *
        52,  6,2, 2,    // 84 Keypad -
    };

    // Spacer panels, again as in KeyboardPanel.
    private static final int spacerPos[] =
    {
        4,  0,  2, 16,
        0,  4, 45,  2,
        36, 0,  2, 16,
        38, 14, 2,  2,
        38, 10, 2,  2,
        44,  0, 2, 16,
    };

    // 127 toggle buttons for the 127 keys.
    private JToggleButton buttons[] = new JToggleButton[127];

    // When initialising, do not respond to events from combo boxes; it may
    // well be that not all controls have been created yet.
    private boolean initialising = true;

    /** Creates new form CustomizePanel */
    public Customize1kPanel(byte [] r)
    {
        rom = r;
        initComponents();

        // Each combo box is paralleled by a spinner, and the two work in 
        // sync.

        spinRemapFrom.setModel(new SpinnerListModel(scanCodes3));
        spinRemapTo.setModel(new SpinnerListModel(scanCodes2));

        // Set up the keys panel. Once more, as in KeyboardPanel, but this
        // time the number shown on a key is its Set 3 scancode, not its
        // internal key number.
        keysPanel.setLayout(new GridBagLayout());
        for (int key = 0; key < (keyPos.length / 4); key++)
        {
            String s = String.format("%02x", 1 + key);
            if (key == 1) s = "83";
            else if (key == 0x7E) s = "84";
            buttons[key] = new JToggleButton(s);
            buttons[key].addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent evt)
                {
                    keyButtonClicked(evt);
                }
            });

            GridBagConstraints constraints = new GridBagConstraints();

            constraints.gridx      = keyPos[key * 4];
            constraints.gridy      = keyPos[key * 4 + 1];
            constraints.gridwidth  = keyPos[key * 4 + 2];
            constraints.gridheight = keyPos[key * 4 + 3];
            constraints.fill       = GridBagConstraints.HORIZONTAL;

            buttons[key].setPreferredSize(null);
            keysPanel.add(buttons[key], constraints);
        }
        // Add the spacers.
        for (int spacer = 0; spacer < (spacerPos.length / 4); spacer++)
        {
            GridBagConstraints constraints = new GridBagConstraints();

            constraints.gridx      = spacerPos[spacer * 4];
            constraints.gridy      = spacerPos[spacer * 4 + 1];
            constraints.gridwidth  = spacerPos[spacer * 4 + 2];
            constraints.gridheight = spacerPos[spacer * 4 + 3];
            constraints.fill       = GridBagConstraints.BOTH;

            JPanel p = new JPanel();

            keysPanel.add(p, constraints);
        }
        // And the line of dummy labels along the bottom.
        for (int col = 0; col < 54; col++)
        {
            GridBagConstraints constraints = new GridBagConstraints();

            constraints.gridx      = col;
            constraints.gridy      = 18;
            constraints.gridwidth  = 1;
            constraints.gridheight = 1;
            constraints.fill       = GridBagConstraints.BOTH;

            JLabel lbl = new JLabel("_^_");

            lbl.setForeground(getBackground());
            keysPanel.add(lbl, constraints);
        }
        // Fill the two combo boxes.
        comboRemapTo.removeAllItems();
        for (int n = 0; n < scanCodes2.length; n++)
        {
            comboRemapTo.addItem(scanCodes2[n]);
        }
        comboRemapFrom.removeAllItems();
        for (int n = 0; n < scanCodes3.length; n++)
        {
            comboRemapFrom.addItem(scanCodes3[n]);
        }
        initialising = false;
    }

    // When one of the keys is clicked, select that key in the 'remap from'
    // combo box.
    private void keyButtonClicked(ActionEvent evt)
    {
        for (int n = 0; n < buttons.length; n++)
        {
            if (evt.getSource().equals((buttons[n])))
            {
                comboRemapFrom.setSelectedIndex(n);
            }
        }
    }



    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        jPanel1 = new javax.swing.JPanel();
        comboRemapFrom = new javax.swing.JComboBox();
        jLabel2 = new javax.swing.JLabel();
        jLabel3 = new javax.swing.JLabel();
        comboRemapTo = new javax.swing.JComboBox();
        spinRemapFrom = new javax.swing.JSpinner();
        spinRemapTo = new javax.swing.JSpinner();
        keysPanel = new javax.swing.JPanel();

        jPanel1.setName("jPanel1"); // NOI18N

        javax.swing.GroupLayout jPanel1Layout = new javax.swing.GroupLayout(jPanel1);
        jPanel1.setLayout(jPanel1Layout);
        jPanel1Layout.setHorizontalGroup(
            jPanel1Layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 100, Short.MAX_VALUE)
        );
        jPanel1Layout.setVerticalGroup(
            jPanel1Layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 100, Short.MAX_VALUE)
        );

        setName("Customize ROM"); // NOI18N

        comboRemapFrom.setModel(new javax.swing.DefaultComboBoxModel(new String[] { "Item 1", "Item 2", "Item 3", "Item 4" }));
        comboRemapFrom.setName("comboRemapFrom"); // NOI18N
        comboRemapFrom.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                comboRemapFromActionPerformed(evt);
            }
        });

        org.jdesktop.application.ResourceMap resourceMap = org.jdesktop.application.Application.getInstance(info.seasip.keyboardhost.KeyboardHostApp.class).getContext().getResourceMap(Customize1kPanel.class);
        jLabel2.setText(resourceMap.getString("jLabel2.text")); // NOI18N
        jLabel2.setName("jLabel2"); // NOI18N

        jLabel3.setText(resourceMap.getString("jLabel3.text")); // NOI18N
        jLabel3.setName("jLabel3"); // NOI18N

        comboRemapTo.setModel(new javax.swing.DefaultComboBoxModel(new String[] { "Item 1", "Item 2", "Item 3", "Item 4" }));
        comboRemapTo.setName("comboRemapTo"); // NOI18N
        comboRemapTo.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                comboRemapToActionPerformed(evt);
            }
        });

        spinRemapFrom.setName("spinRemapFrom"); // NOI18N
        spinRemapFrom.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                spinRemapFromStateChanged(evt);
            }
        });

        spinRemapTo.setName("spinRemapTo"); // NOI18N
        spinRemapTo.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                spinRemapToStateChanged(evt);
            }
        });

        keysPanel.setName("keysPanel"); // NOI18N

        javax.swing.GroupLayout keysPanelLayout = new javax.swing.GroupLayout(keysPanel);
        keysPanel.setLayout(keysPanelLayout);
        keysPanelLayout.setHorizontalGroup(
            keysPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 500, Short.MAX_VALUE)
        );
        keysPanelLayout.setVerticalGroup(
            keysPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 100, Short.MAX_VALUE)
        );

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(this);
        this.setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(layout.createSequentialGroup()
                .addContainerGap()
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addComponent(keysPanel, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                    .addGroup(layout.createSequentialGroup()
                        .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.TRAILING)
                            .addComponent(jLabel2)
                            .addComponent(jLabel3))
                        .addGap(18, 18, 18)
                        .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING, false)
                            .addGroup(layout.createSequentialGroup()
                                .addComponent(comboRemapTo, javax.swing.GroupLayout.PREFERRED_SIZE, 194, javax.swing.GroupLayout.PREFERRED_SIZE)
                                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                                .addComponent(spinRemapTo))
                            .addGroup(layout.createSequentialGroup()
                                .addComponent(comboRemapFrom, javax.swing.GroupLayout.PREFERRED_SIZE, 194, javax.swing.GroupLayout.PREFERRED_SIZE)
                                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                                .addComponent(spinRemapFrom, javax.swing.GroupLayout.PREFERRED_SIZE, 194, javax.swing.GroupLayout.PREFERRED_SIZE)))))
                .addContainerGap())
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(layout.createSequentialGroup()
                .addContainerGap()
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(comboRemapFrom, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(spinRemapFrom, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(jLabel2))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(comboRemapTo, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(jLabel3)
                    .addComponent(spinRemapTo, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                .addGap(65, 65, 65)
                .addComponent(keysPanel, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                .addContainerGap(javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))
        );
    }// </editor-fold>//GEN-END:initComponents

    // When the user tabs off a hex field, parse it as hex and (hopefully) save the
    // result.
    // When the user tabs off a hex field, parse it as hex and (hopefully) save the
    // result.
    // If selection is changed on 'remap from', update the buttons and spin
    // controls with the current selection.
    private void comboRemapFromActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_comboRemapFromActionPerformed
        int n = comboRemapFrom.getSelectedIndex();

        if (n >= 0 && n <= 127)
        {
            byte b = rom[0x381 + n];

            comboRemapTo.setSelectedIndex((b & 0xFF));
            spinRemapFrom.setValue(scanCodes3[n]);
            spinRemapTo  .setValue(scanCodes2[b & 0xFF]);

            for (int m = 0; m < buttons.length; m++)
            {
                if (buttons[m] != null)
                {
                    buttons[m].setSelected(n == m);
                }
            }
        }
    }//GEN-LAST:event_comboRemapFromActionPerformed

    // If selection is changed on 'remap to', update the ROM with the new
    // selection.
    private void comboRemapToActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_comboRemapToActionPerformed
        if (initialising)
        {
            return;
        }
        int n = comboRemapFrom.getSelectedIndex();
        int v = comboRemapTo.getSelectedIndex();

        rom[0x381 + n] = (byte)v;
        spinRemapTo  .setValue(scanCodes2[v]);

    }//GEN-LAST:event_comboRemapToActionPerformed

    // Changes to the spinners update the combo boxes.
    private void spinRemapFromStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_spinRemapFromStateChanged
        SpinnerModel model = (SpinnerModel)spinRemapFrom.getModel();

        Object o = (model.getValue());

        for (int n =0 ; n < scanCodes3.length; n++)
        {
            if (scanCodes3[n].equals(o))
            {
                comboRemapFrom.setSelectedIndex(n);
                break;
            }
        }
    }//GEN-LAST:event_spinRemapFromStateChanged

    // Changes to the spinners update the combo boxes.
    private void spinRemapToStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_spinRemapToStateChanged
       SpinnerModel model = (SpinnerModel)spinRemapTo.getModel();

        Object o = (model.getValue());

        for (int n =0 ; n < scanCodes2.length; n++)
        {
            if (scanCodes2[n].equals(o))
            {
                comboRemapTo.setSelectedIndex(n);
                break;
            }
        }        // TODO add your handling code here:
    }//GEN-LAST:event_spinRemapToStateChanged


    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JComboBox comboRemapFrom;
    private javax.swing.JComboBox comboRemapTo;
    private javax.swing.JLabel jLabel2;
    private javax.swing.JLabel jLabel3;
    private javax.swing.JPanel jPanel1;
    private javax.swing.JPanel keysPanel;
    private javax.swing.JSpinner spinRemapFrom;
    private javax.swing.JSpinner spinRemapTo;
    // End of variables declaration//GEN-END:variables

}
