/* KeyboardHost: Copyright 2011 John Elliott <jce@seasip.demon.co.uk>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package info.seasip.keyboardhost;

/** Interface for the emulated microntroller to interact with the outside
 * world. This emulates its I/O lines.
 *
 * @author John Elliott
 */
public interface IOPort
{
    /** Write a microcontroller port.
     *
     * @param port 0 for the bus, 1-7 for numbered ports.
     * @param value The value to write.
     */
    public void writePort(int port, byte value);


    /** Read a microcontroller port.
     *
     * @param port 0 for the bus, 1-7 for numbered ports.
     * @return The state of the selected port.
     */
    public byte readPort(int port);

    /** Write to the bus.
     *
     * @param value The value to write.
     */
    public void writeBus(byte value);

    /** Read the bus.
     *
     * @return The value read.
     */
    public byte readBus();

    /** Read the T0 line.
     *
     * @return The state of the T0 line.
     */
    public boolean readT0();

    /** Read the T1 line.
     *
     * @return The state of the T1 line.
     */
    public boolean readT1();

    /** Read the I line.
     *
     * @return The state of the I line.
     */
    public boolean readI();

    /** Read external memory.
     *
     * @param addr The address at which to read.
     * @return The byte at that address.
     */
    public byte readExtMem(int addr);

    /** Write external memory.
     *
     * @param addr The address at which to write.
     * @param value The value to write.
     */
    public void writeExtMem(int addr, byte value);
}
